# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Benchmark MDO formulations based on scalable disciplines
========================================================

The :mod:`~gemseo.problems.scalable.study` package implements several classes
to benchmark MDO formulations based on scalable disciplines.

The :class:`.ScalabilityStudy` class implements
the concept of scalability study:

1. By instantiating a :class:`.ScalabilityStudy`, the user defines
   the MDO problem in terms of design parameters, objective function and
   constraints.
2. For each discipline, the user adds a dataset stored
   in a :class:`.AbstractFullCache` and select a type of
   :class:`.ScalableModel` to build the :class:`.ScalableDiscipline`
   associated with this discipline.
3. The user adds different optimization strategies, defined in terms
   of both optimization algorithms and MDO formulation.
4. The user adds different scaling strategies, in terms of sizes of
   design parameters, coupling variables and equality and inequality
   constraints. The user can also define a scaling strategies according to
   particular parameters rather than groups of parameters.
5. Lastly, the user executes the :class:`.ScalabilityStudy` and the results
   are written in several files and stored into directories
   in a hierarchical way, where names depends on both MDO formulation,
   scaling strategy and replications when it is necessary. Different kinds
   of files are stored: optimization graphs, dependency matrix plots and
   of course, scalability results by means of a dedicated class:
   :class:`.ScalabilityResult`.

The :class:`.PostScalabilityStudy` class implements the way as the set of
:class:`.ScalabilityResult`-based result files
contained in the study directory are graphically post-processed. This class
provides several methods to easily change graphical properties, notably
the plot labels. It also makes it possible to define a cost function per
MDO formulation, converting the numbers of executions and linearizations
of the different disciplines required by a MDO process in an estimation
of the computational cost associated with what would be a scaled version
of the true problem.

.. warning::

   Comparing MDO formulations in terms of estimated true computational time
   rather than CPU time of the :class:`.ScalabilityStudy` is highly
   recommended.
   Indeed, time is often an obviousness criterion to distinguish between
   MDO formulations having the same performance in terms of distance to the
   optimum: look at our calculation budget and choose the best formulation
   that satisfies this budget, or even saves us time. Thus, it is important
   to carefully define these cost functions.
"""



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Remi Lafage
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#         Francois Gallard : refactoring for v1, May 2016
"""A factory to execute optimization algorithms from their name."""
from __future__ import division, unicode_literals

from gemseo.algos.driver_factory import DriverFactory
from gemseo.algos.opt.opt_lib import OptimizationLibrary


class OptimizersFactory(DriverFactory):
    """Optimization  libraries factory, see DriverFactory."""

    def __init__(self):
        """Test and import optimization library if it is available."""
        super(OptimizersFactory, self).__init__(OptimizationLibrary, "gemseo.algos.opt")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Uncertainty quantification and management.

The package **uncertainty** provides several functionalities
to quantify and manage uncertainties.
Most of them can be used from the dedicated API.

The sub-package **distributions** offers an abstract level for probability distributions
as well as interfaces to the OpenTURNS and SciPy ones.
It is also possible to fit a probability distribution from data
or select the most likely one from a list of candidates.
These distributions can be used to define random variables in a :class:`.ParameterSpace`
before propagating these uncertainties through a system of :class:`.MDODiscipline`,
by means of a :class:`.DOEScenario`.

.. seealso::

    :class:`.OTDistribution`
    :class:`.SPDistribution`
    :class:`.OTDistributionFitter`

The sub-package **sensitivity** offers an abstract level for sensitivity analysis,
as well as concrete features.
These sensitivity analyses compute indices by means of various methods:
correlation measures, Morris technique and Sobol' variance decomposition.
This sub-package is based in particular on OpenTURNS.

.. seealso::

    :class:`.CorrelationAnalysis`
    :class:`.MorrisAnalysis`
    :class:`.SobolAnalysis`

The sub-package **statistics** offers an abstract level for statistics,
as well as parametric and empirical versions.
Empirical statistics are estimated from a :class:`.Dataset`
while parametric statistics are analytical properties of a :class:`.Distribution`
fitted from a :class:`.Dataset`.

.. seealso::

    :class:`.EmpiricalStatistics`
    :class:`.ParametricStatistics`
"""



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Caching module to avoid multiple evaluations of a discipline
************************************************************
"""
from __future__ import division, unicode_literals

import logging
import sys
from hashlib import sha1
from numbers import Number

from numpy import (
    append,
    array,
    array_equiv,
    ascontiguousarray,
    complex128,
    concatenate,
    float64,
    uint8,
    vstack,
)
from numpy.linalg import norm

from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.ggobi_export import save_data_arrays_to_xml
from gemseo.utils.locks import synchronized, synchronized_hashes
from gemseo.utils.multi_processing import Manager, RLock, Value

TYPE_ERR_MSG = "__getitem__ uses one of these argument types: "
TYPE_ERR_MSG += "int, str, "
TYPE_ERR_MSG += "list(int), list(str), "
TYPE_ERR_MSG += "(int, str), (int, list(str)), "
TYPE_ERR_MSG += "(list(int), str), (int, list(str)) "
TYPE_ERR_MSG += "or (list(int), list(str)). "

LOGGER = logging.getLogger(__name__)


class AbstractCache(object):
    """Abstract class for caches: Defines the common methods for caching inputs,
    outputs, and jacobians of a MDODiscipline.

    See also
    --------
    :class:`.SimpleCache`: store the last evaluation
    :class:`.MemoryFullCache`: store all data in memory
    :class:`.HDF5Cache`: store all data in an HDF5 file
    """

    SAMPLE_GROUP = "sample"
    INPUTS_GROUP = "inputs"
    OUTPUTS_GROUP = "outputs"
    JACOBIAN_GROUP = "jacobian"

    def __init__(self, tolerance=0.0, name=None):
        """Initialize cache tolerance. By default, don't use approximate cache. It is up
        to the user to choose to optimize CPU time with this or not.

        could be something like 2 * finfo(float).eps

        Parameters
        ----------
        tolerance : float
            Tolerance that defines if two input vectors
            are equal and cached data shall be returned.
            If 0, no approximation is made. Default: 0.
        name : str
            Name of the cache.
        """
        self.tolerance = tolerance
        self.name = name if name is not None else self.__class__.__name__
        self._varsizes = None
        self.__inputs_names = None
        self.__outputs_names = None

    def __bool__(self):
        """returns True is the cache is not empty."""
        return len(self) > 0

    @property
    def inputs_names(self):
        """Return the inputs names."""
        if self.__inputs_names is None:
            if not self:
                return None
            data1 = self.get_data(1)[self.INPUTS_GROUP]
            self.__inputs_names = list(iter(data1.keys()))
        return self.__inputs_names

    @property
    def outputs_names(self):
        """Return the outputs names."""
        if self.__outputs_names is None:
            if not self:
                return None
            data1 = self.get_data(1)[self.OUTPUTS_GROUP]
            self.__outputs_names = list(iter(data1.keys()))
        return self.__outputs_names

    @property
    def varsizes(self):
        """Return the variables sizes."""

        def length(obj):
            """Length of an object if __len__ attribute exists, else 1."""
            return len(obj) if hasattr(obj, "__len__") else 1

        if self._varsizes is None:
            inputs = self.get_last_cached_inputs()
            outputs = self.get_last_cached_outputs()
            if inputs is not None and outputs is not None:
                self._varsizes = {key: length(val) for key, val in inputs.items()}
                self._varsizes.update(
                    {key: length(val) for key, val in outputs.items()}
                )
        return self._varsizes

    def __str__(self):
        msg = "Name: {}\n".format(self.name)
        msg += "Type: {}\n".format(self.__class__.__name__)
        msg += "Tolerance: {}\n".format(self.tolerance)
        msg += "Input names: {}\n".format(self.inputs_names)
        msg += "Output names: {}\n".format(self.outputs_names)
        msg += "Length: {}".format(len(self))
        return msg

    def __len__(self):
        return self.get_length()

    def cache_outputs(self, input_data, input_names, output_data, output_names=None):
        """Cache data to avoid re evaluation.

        Parameters
        ----------
        input_data : dict
            Input data to cache.
        input_names : list(str)
            List of input data names.
        output_data : dict
            Output data to cache.
        output_names : list(str)
            List of output data names. If None, use all output names.
            Default: None.
        """
        raise NotImplementedError()

    def cache_jacobian(self, input_data, input_names, jacobian):
        """Cache jacobian data to avoid re evaluation.

        Parameters
        ----------
        input_data : dict
            Input data to cache.
        input_names : list(str)
            List of input data names.
        jacobian : dict
            Jacobian to cache.
        """
        raise NotImplementedError()

    def get_outputs(self, input_data, input_names=None):
        """Check if the discipline has already been evaluated for the given input data
        dictionary. If True, return the associated cache, otherwise return None.

        Parameters
        ----------
        input_data : dict
            Input data dictionary to test for caching.
        input_names : list(str)
            List of input data names.
            If None, takes them all

        Returns
        -------
        output_data : dict
            Output data if there is no need to evaluate the discipline.
            None otherwise.
        jacobian : dict
            Jacobian if there is no need to evaluate the discipline.
            None otherwise.
        """
        raise NotImplementedError()

    def clear(self):
        """Clear the cache."""
        self.__inputs_names = None
        self.__outputs_names = None
        self._varsizes = None

    def get_last_cached_inputs(self):
        """Retrieve the last execution inputs.

        Returns
        -------
        inputs : dict
            Last cached inputs.
        """
        raise NotImplementedError()

    def get_last_cached_outputs(self):
        """Retrieve the last execution outputs.

        Returns
        -------
        outputs : dict
            Last cached outputs.
        """
        raise NotImplementedError()

    def get_all_data(self, **options):
        """Read all the data in the cache.

        Returns
        -------
        all_data : dict
            A dictionary of dictionaries for inputs, outputs and jacobian
            where keys are data indices.
        """
        raise NotImplementedError()

    def get_length(self):
        """Get the length of the cache, ie the number of stored elements.

        Returns
        -------
        length : int
            Length of the cache.
        """
        raise NotImplementedError()

    @property
    def max_length(self):
        """Get the maximal length of the cache (the maximal number of stored elements).

        Returns
        -------
        length : int
            Maximal length of the cache.
        """
        raise NotImplementedError()

    def _check_index(self, index):
        """Check if index is a sample index.

        :param int index: index.
        """
        is_integer = isinstance(index, Number)
        if not is_integer or index < 1 or index > self.get_length():
            raise ValueError(str(index) + " is not a sample index.")

    def get_data(self, index, **options):
        """Returns an elementary sample.

        :param index: sample index.
        :type index: int
        :param options: getter options
        """
        raise NotImplementedError()

    @property
    def samples_indices(self):
        """List of samples indices."""
        return range(1, self.get_length() + 1)


class AbstractFullCache(AbstractCache):
    """Abstract cache to store all data, either in memory or on the disk.

    See also
    --------
    :class:`.MemoryFullCache`: store all data in memory
    :class:`.HDF5Cache`: store all data in an HDF5 file
    """

    def __init__(self, tolerance=0.0, name=None):
        """Initialize cache tolerance. By default, don't use approximate cache. It is up
        to the user to choose to optimize CPU time with this or not.

        could be something like 2 * finfo(float).eps

        Parameters
        ----------
        tolerance : float
            Tolerance that defines if two input vectors
            are equal and cached data shall be returned.
            If 0, no approximation is made. Default: 0.
        name : str
            Name of the cache.
        """
        super(AbstractFullCache, self).__init__(tolerance, name)
        self.lock_hashes = RLock()
        self._manager = Manager()
        self._hashes = self._manager.dict()
        # Maximum index of the data stored in the subgroup of the node
        self._max_group = Value("i", 0)
        self._last_accessed_group = Value("i", 0)
        self.lock = self._set_lock()

    def _set_lock(self):
        """Sets a lock for multithreading, either from an external object or internally
        by using RLock()."""
        raise NotImplementedError

    def __get_or_increment_group_num(self, input_data, data_hash):
        """This method is the second step of caching new inputs. Either gets the right
        cache group number if the inputs are already cached or creates a new one if
        needed.

        :param dict input_data: the input data to cache
        :param int data_hash: the hash of the data
        :return: True if a group was created
        :rtype: bool
        """

        hash_groups = self._hashes.get(data_hash)
        if hash_groups is None:  # Checks if this hash is already stored
            self._max_group.value += 1
            self._last_accessed_group.value = self._max_group.value
            self._hashes[data_hash] = array([self._max_group.value])
            self._initialize_entry(self._max_group.value)
            return True

        # If yes, append to the group
        for group in hash_groups:
            read_in_data = self._read_data(group, self.INPUTS_GROUP)
            if check_cache_equal(input_data, read_in_data):
                # Data is already cached !
                # We dont store it again
                self._last_accessed_group.value = group
                return False

        # Inputs not yet cached, but hashes are equal
        # we need to update the group
        self._max_group.value += 1
        self._last_accessed_group.value = self._max_group.value
        self._hashes[data_hash] = append(hash_groups, self._max_group.value)
        self._initialize_entry(self._max_group.value)
        return True

    def _initialize_entry(self, sample_id):
        """Initialize an entry of the dataset if needed.

        :param int sample_id: sample ID.
        """

    def _has_group(self, sample_id, var_group):
        """Checks if the dataset has the particular variables group filled in.

        :param int sample_id: sample ID.
        :param str var_group: name of the variables group.
        :return: True if the variables group is filled in.
        :rtype: bool
        """
        raise NotImplementedError()

    def _write_data(self, values, names, var_group, sample_id):
        """Writes data associated with a variables group and a sample ID into the
        dataset.

        :param dict values: data dictionary where keys are variables names
            and values are variables values (numpy arrays).
        :param list(str) names: list of input data names to write.
        :param str var_group: name of the variables group,
            either AbstractCache.INPUTS_GROUP, AbstractCache.OUTPUTS_GROUP or
            AbstractCache.JACOBIAN_GROUP.
        :param int sample_id: sample ID.
        """
        raise NotImplementedError()

    def _cache_inputs(self, input_data, input_names, out_group_to_check):
        """This method is the first step step of caching new inputs or Jacobian. It
        caches inputs and increments group if needed. Checks if the out_group_to_check
        exists for these inputs.

        :param dict input_data: data dictionary where keys are variables names
            and values are variables values (numpy arrays).
        :param list(str) input_names: list of input data names
        :param out_group_to_check: name of the variables group to check
            existence, either AbstractCache.OUTPUTS_GROUP or
            AbstractCache.JACOBIAN_GROUP.
        :return: True if the data group exists, avoids duplicate storage.
        :rtype: bool
        """
        data_hash = hash_data_dict(input_data, input_names)
        new_grp = self.__get_or_increment_group_num(input_data, data_hash)

        write_inputs = True
        if not new_grp:
            # The Jacobian may have been stored before but not the outputs yet
            # If the data is already stored, returns
            # Otherwise, store outputs but dont write inputs again
            group_number = self._last_accessed_group.value
            if self._has_group(group_number, out_group_to_check):
                return True
            # Inputs are cached
            write_inputs = False

        if write_inputs:
            self._write_data(
                input_data, input_names, self.INPUTS_GROUP, self._max_group.value
            )

        return False

    @synchronized
    def cache_outputs(self, input_data, input_names, output_data, output_names=None):
        """Cache data to avoid re evaluation.

        Parameters
        ----------
        input_data : dict
            Input data to cache.
        input_names : list(str)
            List of input data names.
        output_data : dict
            Output data to cache.
        output_names : list(str)
            List of output data names. If None, use all output names.
            Default: None.
        """
        data_exists = self._cache_inputs(input_data, input_names, self.OUTPUTS_GROUP)

        if data_exists:
            return

        data_to_cache = output_names or output_data.keys()

        self._write_data(
            output_data, data_to_cache, self.OUTPUTS_GROUP, self._max_group.value
        )

    @synchronized
    def cache_jacobian(self, input_data, input_names, jacobian):
        """Cache jacobian data to avoid re evaluation.

        Parameters
        ----------
        input_data : dict
            Input data to cache.
        input_names : list(str)
            List of input data names.
        jacobian : dict
            Jacobian to cache.
        """
        data_exists = self._cache_inputs(input_data, input_names, self.JACOBIAN_GROUP)
        if data_exists:
            return

        flat_jac = DataConversion.dict_jac_to_dict(jacobian)

        self._write_data(
            flat_jac,
            flat_jac.keys(),
            self.JACOBIAN_GROUP,
            self._last_accessed_group.value,
        )

        return

    @synchronized
    def clear(self):
        """Clears the cache."""
        super(AbstractFullCache, self).clear()
        self._hashes = self._manager.dict()
        self._max_group.value = 0
        self._last_accessed_group.value = 0

    @synchronized
    def get_last_cached_inputs(self):
        """Retrieve the last execution inputs.

        Returns
        -------
        inputs : dict
            Last cached inputs.
        """
        if self._max_group.value == 0:
            return None
        return self._read_data(self._last_accessed_group.value, self.INPUTS_GROUP)

    @synchronized
    def get_last_cached_outputs(self):
        """Retrieve the last execution outputs.

        Returns
        -------
        outputs : dict
            Last cached outputs.
        """
        if self._max_group.value == 0:
            return None
        return self._read_data(self._last_accessed_group.value, self.OUTPUTS_GROUP)

    @synchronized
    def get_length(self):
        """Get the length of the cache, ie the number of stored elements.

        Returns
        -------
        length : int
            Length of the cache.
        """
        return self._max_group.value

    @property
    def max_length(self):
        """Get the maximal length of the cache (the maximal number of stored elements).

        Returns
        -------
        length : int
            Maximal length of the cache.
        """
        return sys.maxsize

    def _read_data(self, group_number, group_name, **options):
        """Read data from a data provider defined in the overloaded classes.

        :param str group_name: name of the group where data is written.
        :param int group_number: number of the group.
        :param options: options passed to the overloaded methods.
        :returns: data dict and jacobian
        """
        raise NotImplementedError()

    @synchronized_hashes
    def __has_hash(self, data_hash):
        """Get data hash.

        :param int data_hash: the hash of the data.
        :return: the data hash.
        """
        out = self._hashes.get(data_hash)
        return out

    def _read_group(self, group_nums, input_data):
        """Read output data and jacobian associated to given input data inside given
        groups.

        :param list(int) group_nums: group numbers.
        :param dict input_data: input data dictionary.
        :returns: The output data and jacobian
            if there is no need to evaluate the
            discipline, None otherwise.
        :rtype: dict, dict
        """
        for group in group_nums:
            read_in_data = self._read_data(group, self.INPUTS_GROUP)
            if check_cache_equal(input_data, read_in_data):
                out_data = self._read_data(group, self.OUTPUTS_GROUP)
                jac = self._read_data(group, self.JACOBIAN_GROUP)
                # self._last_accessed_group.value = group
                return out_data, jac
        return None, None

    @synchronized
    def get_outputs(self, input_data, input_names=None):
        """Check if the discipline has already been evaluated for the given input data
        dictionary. If True, return the associated cache, otherwise return None.

        Parameters
        ----------
        input_data : dict
            Input data dictionary to test for caching.
        input_names : list(str)
            List of input data names.

        Returns
        -------
        output_data : dict
            Output data if there is no need to evaluate the discipline.
            None otherwise.
        jacobian : dict
            Jacobian if there is no need to evaluate the discipline.
            None otherwise.
        """
        if self.tolerance == 0.0:
            data_hash = hash_data_dict(input_data, input_names)
            group_nums = self.__has_hash(data_hash)
            if group_nums is None:
                return None, None
            return self._read_group(group_nums, input_data)
        for group_nums in self._hashes.values():
            for group in group_nums:
                read_in_data = self._read_data(group, self.INPUTS_GROUP)
                if check_cache_approx(input_data, read_in_data, self.tolerance):
                    out_data = self._read_data(group, self.OUTPUTS_GROUP)
                    jac = self._read_data(group, self.JACOBIAN_GROUP)
                    # self._last_accessed_group.value = group
                    return out_data, jac
        return None, None

    @property
    def _all_groups(self):
        """Sorted group numbers."""
        tmp = []
        for group_nums in self._hashes.values():
            tmp += list(group_nums)
        return sorted(tmp)

    @synchronized
    def _get_all_data(self):
        """Same as _all_data() but with pre- and post- treatment."""
        for data in self._all_data():
            yield data

    @synchronized
    def _all_data(self, **options):
        """Iterator of all data in the cache.

        :returns: sample ID, input data, output data and jacobian.
        :rtype: dict
        """
        for group in self._all_groups:
            read_in_data = self._read_data(group, self.INPUTS_GROUP, **options)
            out_data = self._read_data(group, self.OUTPUTS_GROUP, **options)
            jac = self._read_data(group, self.JACOBIAN_GROUP, **options)
            yield {
                self.SAMPLE_GROUP: group,
                self.INPUTS_GROUP: read_in_data,
                self.OUTPUTS_GROUP: out_data,
                self.JACOBIAN_GROUP: jac,
            }

    @synchronized
    def get_data(self, index, **options):
        """Gets the data associated to a sample ID.

        :param str index: sample ID.
        :param options: options passed to the _read_data() method.
        :return: input data, output data and jacobian.
        :rtype: dict
        """
        self._check_index(index)
        input_data = self._read_data(index, self.INPUTS_GROUP, **options)
        output_data = self._read_data(index, self.OUTPUTS_GROUP, **options)
        jacobian = self._read_data(index, self.JACOBIAN_GROUP, **options)
        return {
            self.INPUTS_GROUP: input_data,
            self.OUTPUTS_GROUP: output_data,
            self.JACOBIAN_GROUP: jacobian,
        }

    @synchronized
    def get_all_data(self, as_iterator=False):  # pylint: disable=W0221
        """Return all the data in the cache.

        Parameters
        ----------
        as_iterator : bool
            If True, return an iterator. Otherwise a dictionary.
            Default: False.

        Returns
        -------
        all_data : dict
            A dictionary of dictionaries for inputs, outputs and jacobian
            where keys are data indices.
        """
        if as_iterator:
            return self._get_all_data()

        all_data = {}
        for data in self._get_all_data():
            sample = data[self.SAMPLE_GROUP]
            inputs = data[self.INPUTS_GROUP]
            outputs = data[self.OUTPUTS_GROUP]
            jac = data[self.JACOBIAN_GROUP]
            all_data[sample] = {
                self.INPUTS_GROUP: inputs,
                self.OUTPUTS_GROUP: outputs,
                self.JACOBIAN_GROUP: jac,
            }
        return all_data

    def export_to_ggobi(self, file_path, inputs_names=None, outputs_names=None):
        """Export history to xml file format for ggobi tool.

        Parameters
        ----------
        file_path : str
            Path to export the file.
        inputs_names : list(str)
            List of inputs to include in the export.
            By default, take all of them.
        outputs_names : list(str)
            Names of outputs to export.
            By default, take all of them.
        """
        n_groups = len(self._hashes)
        if n_groups == 0:
            raise ValueError("Empty cache!")
        in_n = None
        out_n = None
        in_d = []
        out_d = []
        len_dict = {}

        for data in self.get_all_data(True):
            in_dict = data[self.INPUTS_GROUP] or {}
            out_dict = data[self.OUTPUTS_GROUP] or {}
            try:
                if inputs_names is not None:
                    in_dict = {key: in_dict[key] for key in inputs_names}
                if outputs_names is not None:
                    out_dict = {key: out_dict[key] for key in outputs_names}
            except KeyError:
                # The data is not in this execution
                continue
            # Compute the size of the data
            len_dict.update({key: val.size for key, val in in_dict.items()})
            len_dict.update({key: val.size for key, val in out_dict.items()})
            in_keys = set(in_dict.keys())
            out_keys = set(out_dict.keys())
            in_n = (in_n or in_keys) & in_keys
            out_n = (out_n or out_keys) & out_keys
            in_d.append(in_dict)
            out_d.append(out_dict)

        if not out_d:
            raise ValueError("Failed to find outputs in the cache")

        variables_names = []
        for data_name in list(in_n) + list(out_n):
            len_data = len_dict[data_name]
            if len_data == 1:
                variables_names.append(data_name)
            else:
                variables_names += [
                    data_name + "_" + str(i + 1) for i in range(len_data)
                ]
        values_list = (
            concatenate(
                [in_d[g][i].flatten() for i in in_n]
                + [out_d[g][o].flatten() for o in out_n]
            )
            for g in range(len(in_d))
        )
        values_array = vstack(values_list)
        save_data_arrays_to_xml(variables_names, values_array, file_path)

    def merge(self, other_cache):
        """Merges an other cache with self.

        Parameters
        ----------
        other_cache : AbstractFullCache
            Cache to merge with the current one.
        """
        for data in other_cache.get_all_data(True):
            input_data = data[self.INPUTS_GROUP]
            output_data = data[self.OUTPUTS_GROUP]
            jacobian = data[self.JACOBIAN_GROUP]
            if other_cache.inputs_names is not None:
                inputs_names = other_cache.inputs_names
            else:
                inputs_names = self.inputs_names
            outputs_names = other_cache.outputs_names or self.outputs_names
            self.cache_outputs(input_data, inputs_names, output_data, outputs_names)
            if jacobian is not None:
                self.cache_jacobian(input_data, inputs_names, jacobian)

    def _duplicate_from_scratch(self):
        """Duplicate a cache from scratch, ie.

        only duplicate the construction step.
        """
        raise NotImplementedError()

    def __add__(self, other_cache):
        """Add another cache to the current cache and returns the sum. dtype="bytes"))

            for data_name in data_names:
                val = data.get(data_name)
                if val is not None:
                    io_group.create_dataset(data_name, data=to_real(val))
        except RuntimeError as err:
            h5file.close()
            node_path = str(hdf_node_path) + "." + str(group_num)
            node_path += "." + str(group_name)
            raise RuntimeError("Failed to cache dataset " + node_path +
                               " in file : " + str(self.hdf_file_path) +
                               " h5py error : " + str(err.args[0]))
        data = {key: array(val) for key, val in values_group.items()}
            data_hash = int(array(read_hash)[0])

        :param AbstractCache other_cache: other cache
        :return: cache with the same class as the first cache in the sum.
        """
        new_cache = self._duplicate_from_scratch()
        new_cache.merge(self)
        new_cache.merge(other_cache)
        return new_cache

    def export_to_dataset(
        self,
        name=None,
        by_group=True,
        categorize=True,
        inputs_names=None,
        outputs_names=None,
    ):
        """Set Dataset from a cache.

        :param str name: dataset name.
        :param bool by_group: if True, store the data by group. Otherwise,
            store them by variables. Default: True
        :param bool categorize: distinguish between the different groups of
            variables. Default: True.
        :param list(str) inputs_names: list of inputs names. If None, use all
            inputs. Default: None.
        :param list(str) outputs_names: list of outputs names. If None, use all
            outputs. Default: None.
        """
        from gemseo.core.dataset import Dataset

        dataset = Dataset(name or self.name, by_group)

        to_array = DataConversion.list_of_dict_to_array

        inputs_names = inputs_names or self.inputs_names
        outputs_names = outputs_names or self.outputs_names

        # Set the different groups
        in_grp = out_grp = dataset.DEFAULT_GROUP
        cache_output_as_input = True
        if categorize:
            in_grp = dataset.INPUT_GROUP
            out_grp = dataset.OUTPUT_GROUP
            cache_output_as_input = False

        # Add cache inputs and outputs
        data = list(self.get_all_data(True))
        inputs = to_array(data, inputs_names, self.INPUTS_GROUP)
        data = DataConversion.array_to_dict(inputs, inputs_names, self.varsizes)
        for input_name, value in data.items():
            dataset.add_variable(input_name, value, in_grp)
        data = list(self.get_all_data(True))
        outputs = to_array(data, outputs_names, self.OUTPUTS_GROUP)
        data = DataConversion.array_to_dict(outputs, outputs_names, self.varsizes)
        for output_name, value in data.items():
            dataset.add_variable(
                output_name, value, out_grp, cache_as_input=cache_output_as_input
            )
        return dataset


def hash_data_dict(data, names_tokeep=None):
    """Hash a data dict using sha1.
        for group_num, group in node_group.items():
            hash_value = int(array(read_hash)[0])

    Parameters
    ----------
    data : dict
        The data dictionary.
    names_tokeep : list(str)
        Names of the data to keep for hashing.
        If None, use sorted(data.keys()).

    Returns
    -------
    hash : int
        Hash value of the data dictionary.

    Examples
    --------
    >>> from gemseo.core.cache import hash_data_dict
    >>> from numpy import array
    >>> data = {'x':array([1.,2.]),'y':array([3.])}
    >>> hash_data_dict(data)
    1871784392126344814771968055738742895695521374568L
    >>> hash_data_dict(data,'x')
    756520441774735697349528776513537427923146459919L
    """
    hash_val = 0
    iterator = names_tokeep
    if names_tokeep is None:
        iterator = iter(data.keys())
    for name in iterator:
        val = data.get(name)
        if val is None:
            continue
        # We sum all the hashes so that the hash does not
        # depend on the order of the keys
        try:
            hash_val += int(sha1(val.view(uint8)).hexdigest(), 16)
        except ValueError:  # View may not support discontiguous arrays
            sha1val = sha1(ascontiguousarray(val).view(uint8))
            hash_val += int(sha1val.hexdigest(), 16)
    return hash_val


def check_cache_approx(data_dict, cache_dict, cache_tol=0.0):
    """Checks if the data_dict is approximately equal to the cache dict at
    self.tolerance (absolute + relative)

    :param data_dict: data dict to check
    :returns: True if the dict are approximately equal
    """
    for key, val in cache_dict.items():
        norm_err = norm((data_dict[key] - val))
        if norm_err > cache_tol * (1.0 + norm(val)):
            return False

    return True


def check_cache_equal(data_dict, cache_dict):
    """Check if the data dictionary is equal to the cache data dictionary.

    Parameters
    ----------
    data_dict : dict
        Data dictionary to check.
    cache_dict : dict
        Cache data dictionary to check.

    Returns
    -------
    is_equal : bool
        True if the dictionaries are equal.

    Examples
    --------
    >>> from numpy import array
    >>> data_1 = {'x': array([1.]), 'y': array([2.])}
    >>> data_2 = {'x': array([1.]), 'y': array([3.])}
    >>> check_cache_equal(data_1, data_1)
    True
    >>> check_cache_equal(data_1, data_2)
    False
    """
    for key, buff_val in cache_dict.items():
        in_val = data_dict.get(key)
        if in_val is None or not array_equiv(in_val, buff_val):
            return False
    return True


def to_real(data):
    """Convert complex to real numpy array."""
    if data.dtype == complex128:
        return array(array(data, copy=False).real, dtype=float64)
    return data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Draw a radar chart from a :class:`.Dataset`. """
from __future__ import division, unicode_literals

from typing import List, Mapping

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from numpy import linspace, pi, rad2deg

from gemseo.post.dataset.dataset_plot import DatasetPlot


class RadarChart(DatasetPlot):
    """Radar Chart visualization."""

    def _plot(
        self,
        properties,  # type: Mapping
        display_zero=True,  # type: bool
        connect=False,  # type: bool
        radial_ticks=False,  # type: bool
    ):  # type: (...) -> List[Figure]
        """
        Args:
            display_zero: If True, display the line where the output is equal to zero.
            connect: If True, connect the elements of a series with a line.
            radial_ticks: If True, align the ticks names with the radius.
        """
        linestyle = "-o" if connect else "o"

        fig = plt.figure(figsize=self.figsize)
        axe = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection="polar")
        axe.grid(True, color="k", linewidth=0.3, linestyle=":")
        axe.tick_params(labelsize=self.font_size)

        all_data, _, sizes = self.dataset.get_all_data(False, False)
        variables_names = self.dataset.columns_names
        if self.rmin is None:
            self.rmin = all_data.min()
        if self.rmax is None:
            self.rmax = all_data.max()
        dimension = sum(sizes.values())

        # computes angles
        theta = (2 * pi * linspace(0, 1 - 1.0 / dimension, dimension)).tolist()
        theta.append(theta[0])

        series_names = self.dataset.row_names
        if self.color is None:
            colormap = plt.cm.get_cmap(self.colormap)
            self.color = {
                name: colormap(color)
                for name, color in zip(series_names, linspace(0, 1, len(all_data)))
            }
        if self.linestyle is None:
            self.linestyle = {name: linestyle for name in series_names}
        for index, data in enumerate(all_data):
            name = series_names[index]
            data = data.tolist()
            data.append(data[0])
            axe.plot(
                theta,
                data,
                self.linestyle[name],
                color=self.color[name],
                lw=1,
                label=name,
            )
        if display_zero and self.rmin < 0:
            circle = plt.Circle(
                (0, 0),
                abs(self.rmin),
                transform=axe.transData._b,
                fill=False,
                edgecolor="black",
                linewidth=1,
                zorder=10,
            )
            plt.gca().add_artist(circle)
        theta_degree = rad2deg(theta[:-1])
        axe.set_thetagrids(theta_degree, variables_names)
        if radial_ticks:
            labels = []
            for label, angle in zip(axe.get_xticklabels(), theta_degree):
                x, y = label.get_position()
                lab = axe.text(
                    x,
                    y,
                    label.get_text(),
                    transform=label.get_transform(),
                    ha=label.get_ha(),
                    va=label.get_va(),
                )
                if 90 < angle <= 180:
                    angle = 360 - (180 - angle)
                if 180 < angle < 270:
                    angle = angle - 180
                lab.set_rotation(angle)
                labels.append(lab)
            axe.set_xticklabels([])
        axe.set_rlim([self.rmin, self.rmax])
        axe.set_rticks(linspace(self.rmin, self.rmax, 6))
        axe.legend(
            loc="upper left", fontsize=self.font_size, bbox_to_anchor=(1.05, 1.0)
        )
        axe.set_title(self.title, fontsize=self.font_size * 1.2)
        box = axe.get_position()
        axe.set_position(
            [box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9]
        )
        axe.legend(
            loc="upper center",
            bbox_to_anchor=(0.5, -0.05),
            ncol=5,
        )
        return [fig]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret, Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A set of Newton algorithm variants for solving MDAs
***************************************************

Root finding methods include:

- `Newton-Raphson <https://en.wikipedia.org/wiki/Newton%27s_method>`__
- `quasi-Newton methods <https://en.wikipedia.org/wiki/Quasi-Newton_method>`__

Each of these methods is implemented by a class in this module.
Both inherits from a common abstract cache.
"""
from __future__ import division, unicode_literals

import logging
from copy import deepcopy

from numpy.linalg import norm
from scipy.optimize import root

from gemseo.core.discipline import MDODiscipline
from gemseo.mda.mda import MDA
from gemseo.utils.data_conversion import DataConversion

# from gemseo.core.parallel_execution import DisciplinesParallelExecution
LOGGER = logging.getLogger(__name__)


class MDARoot(MDA):
    """Abstract class implementing MDAs based on (Quasi-)Newton methods."""

    def __init__(
        self,
        disciplines,
        max_mda_iter=10,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param name: the name of the chain
        :type name: str
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        """
        self.tolerance = 1e-6
        self.max_mda_iter = 10
        super(MDARoot, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            grammar_type=grammar_type,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
            use_lu_fact=use_lu_fact,
        )
        self._initialize_grammars()
        self._set_default_inputs()
        self._compute_input_couplings()
        # parallel execution
        # ==================================================================
        # if 1 < n_processes:
        #     self.parallel_execution = DisciplinesParallelExecution(
        #         self.disciplines, n_processes)
        # else:
        #    self.parallel_execution = None
        # ==================================================================

    def _initialize_grammars(self):
        """Define all inputs and outputs of the chain."""
        for disciplines in self.disciplines:
            self.input_grammar.update_from(disciplines.input_grammar)
            self.output_grammar.update_from(disciplines.output_grammar)

    def execute_all_disciplines(self, input_local_data):
        """Execute all self.disciplines.

        :param input_local_data: the input data of the disciplines
        """
        # Set status of sub disciplines
        # if self.parallel_execution is not None:
        #     self.disciplines = self.parallel_execution
        # .execute(input_local_data)
        # else:
        for discipline in self.disciplines:
            discipline.reset_statuses_for_run()
            discipline.execute(deepcopy(input_local_data))

        outputs = [discipline.get_output_data() for discipline in self.disciplines]
        for data in outputs:
            self.local_data.update(data)

    def _run(self):
        """Run the MDA.

        To be implemented in concrete classes.
        """
        raise NotImplementedError()


class MDANewtonRaphson(MDARoot):
    r"""Newton solver for MDA.

    The `Newton-Raphson method
    <https://en.wikipedia.org/wiki/Newton%27s_method>`__ is parameterized by a
    relaxation factor :math:`\alpha \in (0, 1]` to limit the length of the
    steps taken along the Newton direction.  The new iterate is given by:

    .. math::

       x_{k+1} = x_k - \alpha f'(x_k)^{-1} f(x_k)
    """

    def __init__(
        self,
        disciplines,
        max_mda_iter=10,
        relax_factor=0.99,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        linear_solver="lgmres",
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
    ):
        """Constructor.

        :param disciplines: list of disciplines
        :type disciplines: list(MDODiscipline)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param relax_factor: relaxation factor in the Newton step (default 1.)
        :type relax_factor: float
        :param name: name
        :type name: str
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param linear_solver: linear solver used to compute the Newton step
        :type linear_solver: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        """
        super(MDANewtonRaphson, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            grammar_type=grammar_type,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
            use_lu_fact=use_lu_fact,
        )

        self.relax_factor = self.__check_relax_factor(relax_factor)
        self.linear_solver = linear_solver

    @staticmethod
    def __check_relax_factor(relax_factor):
        """Check that the relaxation factor in the Newton step is in (0, 1].

        :param relax_factor: relaxation factor
        """
        if relax_factor <= 0.0 or relax_factor > 1:
            raise ValueError(
                "Newton relaxation factor should belong to "
                + "(0, 1] (current value: "
                + str(relax_factor)
                + ")"
            )
        return relax_factor

    def _newton_step(self):
        """Execute the full newton step.

        Computes increment : -[dR/dW]**-1 . R
        Runs disciplines
        """
        newton_dict = self.assembly.compute_newton_step(
            self.local_data,
            self.strong_couplings,
            self.relax_factor,
            self.linear_solver,
        )
        # update current solution with Newton step
        exec_data = deepcopy(self.local_data)
        for c_var, c_step in newton_dict.items():
            exec_data[c_var] += c_step
        self.reset_disciplines_statuses()
        self.execute_all_disciplines(exec_data)

    def _run(self):
        """Run the MDA.

        Run the disciplines in a sequential way until the difference between
        outputs is under tolerance

        :returns: the local data
        """
        if self.warm_start:
            self._couplings_warm_start()
        # execute the disciplines
        current_couplings = self._current_input_couplings()
        self.reset_disciplines_statuses()
        self.execute_all_disciplines(self.local_data)
        new_couplings = self._current_input_couplings()

        # store initial residual
        current_iter = 1
        self._compute_residual(
            current_couplings, new_couplings, current_iter, first=True
        )
        current_couplings = new_couplings

        while not self._termination(current_iter):
            self._newton_step()
            new_couplings = self._current_input_couplings()

            # store current residual
            current_iter += 1
            self._compute_residual(current_couplings, new_couplings, current_iter)
            current_couplings = new_couplings


class MDAQuasiNewton(MDARoot):
    """Quasi-Newton solver for MDA.

    `Quasi-Newton methods
    <https://en.wikipedia.org/wiki/Quasi-Newton_method>`__
    include numerous variants
    (
    `Broyden
    <https://en.wikipedia.org/wiki/Broyden%27s_method>`__,
    `Levenberg-Marquardt
    <https://en.wikipedia.org/wiki/Levenberg%E2%80%93Marquardt_algorithm>`__,
    ...).
    The name of the variant should be provided
    as a parameter :code:`method` of the class.

    The new iterate is given by:

    .. math::

       x_{k+1} = x_k - \\rho_k B_k f(x_k)

    where :math:`\\rho_k` is a coefficient chosen
    in order to minimize the convergence and :math:`B_k` is an approximation
    of the inverse of the Jacobian :math:`Df(x_k)^{-1}`.
    """

    # quasi-Newton methods
    HYBRID = "hybr"
    LEVENBERG_MARQUARDT = "lm"
    BROYDEN1 = "broyden1"
    BROYDEN2 = "broyden2"
    ANDERSON = "anderson"
    LINEAR_MIXING = "linearmixing"
    DIAG_BROYDEN = "diagbroyden"
    EXCITING_MIXING = "excitingmixing"
    KRYLOV = "krylov"
    DF_SANE = "df-sane"

    QUASI_NEWTON_METHODS = [
        HYBRID,
        LEVENBERG_MARQUARDT,
        BROYDEN1,
        BROYDEN2,
        ANDERSON,
        LINEAR_MIXING,
        DIAG_BROYDEN,
        EXCITING_MIXING,
        KRYLOV,
        DF_SANE,
    ]

    def __init__(
        self,
        disciplines,
        max_mda_iter=10,
        name=None,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        method=HYBRID,
        use_gradient=False,
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param name: name
        :type name: str
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param method: method name in scipy root finding,
            among self.QUASI_NEWTON_METHODS
        :type method: str
        :param use_gradient: if True, used the analytic gradient of
            the discipline
        :type use_gradient: bool
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        """
        super(MDAQuasiNewton, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            grammar_type=grammar_type,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
            use_lu_fact=use_lu_fact,
        )
        if method not in self.QUASI_NEWTON_METHODS:
            msg = "Method " + method + " is not a valid quasi-Newton method"
            raise ValueError(msg)
        self.method = method
        self.use_gradient = use_gradient
        self.local_residual_history = []
        self.last_outputs = None  # used for computing the residual history

    def _solver_options(self):
        """Determine options for the solver, based on the resolution method."""
        options = {}
        if self.method in [
            self.BROYDEN1,
            self.BROYDEN2,
            self.ANDERSON,
            self.LINEAR_MIXING,
            self.DIAG_BROYDEN,
            self.EXCITING_MIXING,
            self.KRYLOV,
        ]:
            options["ftol"] = self.tolerance
            options["maxiter"] = self.max_mda_iter
        elif self.method in [self.LEVENBERG_MARQUARDT]:
            options["xtol"] = self.tolerance
            options["maxiter"] = self.max_mda_iter
        elif self.method in [self.DF_SANE]:
            options["fatol"] = self.tolerance
            options["maxfev"] = self.max_mda_iter
        elif self.method in [self.HYBRID]:
            options["xtol"] = self.tolerance
            options["maxfev"] = self.max_mda_iter
        return options

    def _methods_with_callback(self):
        """Determine whether resolution method accepts a callback function."""
        return [self.BROYDEN1, self.BROYDEN2]

    def _run(self):
        """Run the MDA.

        Runs the disciplines in a sequential way until the difference
        between outputs is under tolerance

        :returns: the local data
        """
        if self.warm_start:
            self._couplings_warm_start()
        self.reset_disciplines_statuses()
        self.execute_all_disciplines(deepcopy(self.local_data))

        couplings = self.strong_couplings

        if not couplings:
            msg = (
                "MDAQuasiNewton found no strong couplings. Executed all"
                "disciplines once."
            )
            LOGGER.warning(msg)
            return self.local_data
        options = self._solver_options()

        def fun(x_vect):
            """Evaluate all residuals, possibly in parallel.

            :param x_vect: design variables
            """
            # transform input vector into a dict
            input_values = DataConversion.update_dict_from_array(
                self.local_data, couplings, x_vect
            )
            # compute all residuals
            self.reset_disciplines_statuses()
            self.execute_all_disciplines(input_values)
            residuals = self.assembly.residuals(input_values, couplings).real
            #             if residuals.size == 1:  # Weak couplings already treated
            #                 return residuals[0]
            return residuals

        jac = None
        if self.use_gradient:

            for discipline in self.disciplines:
                # Tells the discipline what to linearize
                outs = discipline.get_output_data_names()
                to_linearize = set(outs) & set(couplings)
                discipline.add_differentiated_outputs(list(to_linearize))
                inpts = discipline.get_input_data_names()
                to_linearize = set(inpts) & set(couplings)
                discipline.add_differentiated_inputs(list(to_linearize))
            # linearize the residuals

            def jacobian(x_vect):
                """Linearize all the residuals.

                :param x_vect:design variables
                """
                # transform input vector into a dict
                input_values = DataConversion.update_dict_from_array(
                    self.local_data, couplings, x_vect
                )
                # linearize all residuals
                self.reset_disciplines_statuses()
                for discipline in self.disciplines:
                    discipline.linearize(input_values)
                # assemble the system
                n_couplings = 0
                for coupling in couplings:
                    discipline = self.coupling_structure.find_discipline(coupling)
                    size = list(discipline.jac[coupling].values())[0].shape[0]
                    n_couplings += size
                self.assembly.compute_sizes(couplings, couplings, couplings)
                dresiduals = self.assembly.dres_dvar(
                    couplings, couplings, n_couplings, n_couplings
                ).todense()
                return dresiduals

            jac = jacobian

        # initial solution
        y_0 = DataConversion.dict_to_array(self.local_data, couplings).real
        # callback function to retrieve the residual at iteration k
        norm_0 = norm(y_0.real)
        if self.reset_history_each_run:
            self.residual_history = []

        # callback function to store residuals
        self.last_outputs = y_0
        if self.method in self._methods_with_callback():

            def callback(y_k, _):
                """Store the current residual in the history.

                :param y_k: couling variables
                :param _: ignored
                """
                delta = norm((y_k - self.last_outputs).real) / norm_0
                self.residual_history.append((delta, 0))  # iter number?
                self.last_outputs = y_k

        else:
            callback = None

        # solve the system
        y_opt = root(
            fun, x0=y_0, method=self.method, jac=jac, callback=callback, options=options
        )
        # transform optimal vector into a dict
        self.local_data = DataConversion.update_dict_from_array(
            self.local_data, couplings, y_opt.x
        )
        return self.local_data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Benoit Pauwels
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Post-optimal analysis
*********************
"""
from __future__ import division, unicode_literals

import logging

from numpy import atleast_1d, dot, hstack, ndarray, vstack, zeros_like
from numpy.linalg.linalg import norm

from gemseo.algos.lagrange_multipliers import LagrangeMultipliers

LOGGER = logging.getLogger(__name__)


class PostOptimalAnalysis(object):
    r"""
    Post-optimal analysis of a parameterized optimization problem.

    Consider the parameterized optimization problem below, whose objective and
    constraint functions depend on both the optimization variable :math:`x` and
    a parameter :math:`p`.

    .. math::
        \begin{aligned}
            & \text{Minimize}    & & f(x,p) \\
            & \text{relative to} & & x \\
            & \text{subject to}  & & \left\{\begin{aligned}
                                                & g(x,p)\le0, \\
                                                & h(x,p)=0, \\
                                                & \ell\le x\le u.
                                     \end{aligned}\right.
        \end{aligned}

    Denote :math:`x^\ast(p)` a solution of the problem, which depends on
    :math:`p`.
    The post-optimal analysis consists in computing the following total
    derivative:

    .. math::
        \newcommand{\total}{\mathrm{d}}
        \frac{\total f(x^\ast(p),p)}{\total p}(p)
        =\frac{\partial f}{\partial p}(x^\ast(p),p)
         +\lambda_g^\top\frac{\partial g}{\partial p}(x^\ast(p),p)
         +\lambda_h^\top\frac{\partial h}{\partial p}(x^\ast(p),p),

    where :math:`\lambda_g` and :math:`\lambda_h` are the Lagrange multipliers
    of :math:`x^\ast(p)`.
    N.B. the equality above relies on the assumption that

    .. math::
        \newcommand{\total}{\mathrm{d}}
        \lambda_g^\top\frac{\total g(x^\ast(p),p)}{\total p}(p)=0
        \text{ and }
        \lambda_h^\top\frac{\total h(x^\ast(p),p)}{\total p}(p)=0.
    """
    # Dictionary key for term "Lagrange multipliers dot constraints Jacobian"
    MULT_DOT_CONSTR_JAC = "mult_dot_constr_jac"

    def __init__(self, opt_problem, ineq_tol=None):
        """Constructor.

        :param opt_problem: solved optimization problem to be analyzed
        :type opt_problem: OptimizationProblem
        :param ineq_tol: tolerance to determine active inequality constraints.
            If None, its value is fetched in the optimization problem.
        """
        self.lagrange_computer = LagrangeMultipliers(opt_problem)
        # N.B. at creation LagrangeMultipliers checks the optimization problem
        self.opt_problem = opt_problem
        # Get the optimal solution
        self.x_opt = self.opt_problem.design_space.get_current_x()
        # Get the objective name
        outvars = self.opt_problem.objective.outvars
        if len(outvars) != 1:
            raise ValueError("The objective must be single-valued.")
        self.outvars = outvars
        # Set the tolerance on inequality constraints
        if ineq_tol is None:
            self.ineq_tol = self.opt_problem.ineq_tolerance
        else:
            self.ineq_tol = ineq_tol

    def check_validity(self, total_jac, partial_jac, parameters, threshold):
        """Checks whether the assumption for post-optimal validity holds.

        :param total_jac: total derivatives of the post-optimal constraints
        :type total_jac: dict(dict(ndarray))
        :param partial_jac: partial derivatives of the constraints
        :type total_jac: dict(dict(ndarray))
        :param parameters: names list of the optimization problem parameters
        :type parameters: list(str)
        :param threshold: tolerance on the validity assumption
        :type threshold: number
        """
        # Check the Jacobians
        func_names = self.opt_problem.get_constraints_names()
        self._check_jacobians(total_jac, func_names, parameters)
        self._check_jacobians(partial_jac, func_names, parameters)

        # Compute the Lagrange multipliers
        multipliers = self.lagrange_computer.compute(self.x_opt, self.ineq_tol)
        _, mul_ineq = multipliers.get(LagrangeMultipliers.INEQUALITY, ([], []))
        _, mul_eq = multipliers.get(LagrangeMultipliers.EQUALITY, ([], []))

        # Get the array to validate the inequality constraints
        total_ineq_jac = self._get_act_ineq_jac(total_jac, parameters)
        partial_ineq_jac = self._get_act_ineq_jac(partial_jac, parameters)
        ineq_tot, ineq_part, ineq_corr = self._compute_validity(
            total_ineq_jac, partial_ineq_jac, mul_ineq, parameters
        )

        # Get the array to validate the equality constraints
        total_eq_jac = self._get_eq_jac(total_jac, parameters)
        partial_eq_jac = self._get_eq_jac(partial_jac, parameters)
        eq_tot, eq_part, eq_corr = self._compute_validity(
            total_eq_jac, partial_eq_jac, mul_eq, parameters
        )

        # Compute the error
        error_list = [arr for arr in [ineq_tot, eq_tot] if arr is not None]
        part_norm_list = [arr for arr in [ineq_part, eq_part] if arr is not None]
        if error_list and part_norm_list:
            error = norm(vstack(error_list))
            part_norm = norm(vstack(part_norm_list))
            if part_norm > threshold:
                error /= part_norm
        else:
            error = 0.0

        # Assess the validity
        valid = error < threshold
        if valid:
            LOGGER.info("Post-optimality is valid.")
        else:
            msg = "Post-optimality assumption is wrong by "
            msg += str(error * 100.0) + "%."
            LOGGER.info(msg)

        return valid, ineq_corr, eq_corr

    def _compute_validity(self, total_jac, partial_jac, multipliers, parameters):
        """Computes the arrays necessary to the validity check.

        :param total_jac: total derivatives of the post-optimal constraints
        :type total_jac: dict(dict(ndarray))
        :param partial_jac: partial derivatives of the constraints
        :type total_jac: dict(dict(ndarray))
        :param multipliers: Lagrange multipliers
        :type multipliers: ndarray
        :param parameters: names list of the optimization problem parameters
        :type parameters: list(str)
        """
        corrections = dict.fromkeys(parameters, 0.0)  # corrections terms
        total_prod_list = []
        partial_prod_list = []
        for input_name in parameters:
            total_jac_block = total_jac.get(input_name)
            partial_jac_block = partial_jac.get(input_name)
            if total_jac_block is not None and partial_jac_block is not None:
                total_prod_block = dot(multipliers, total_jac_block)
                partial_prod_block = dot(multipliers, partial_jac_block)
                total_prod_list.append(total_prod_block)
                partial_prod_list.append(partial_prod_block)
                # Store the correction term
                corrections[input_name] = -total_prod_block
                if not self.opt_problem.minimize_objective:
                    corrections[input_name] *= -1.0
        total_prod = hstack(total_prod_list) if total_prod_list else None
        partial_prod = hstack(partial_prod_list) if partial_prod_list else None

        return total_prod, partial_prod, corrections

    def execute(self, outputs, inputs, functions_jac):
        """Performs the post-optimal analysis.

        :param outputs: names list of the outputs to differentiate
        :type outputs: list(str)
        :param inputs: names list of the inputs w.r.t. which to differentiate
        :type inputs: list(str)
        :param functions_jac: Jacobians of the optimization functions w.r.t.
            the differentiation inputs
        :type functions_jac: dict(dict(ndarray))
        """
        # Check the outputs
        nondifferentiable_outputs = set(outputs) - set(self.outvars)
        if nondifferentiable_outputs:
            raise ValueError(
                "Only the post-optimal Jacobian of "
                + self.outvars[0]
                + " can be computed"
                + ", not the one(s) of "
                + ", ".join(nondifferentiable_outputs)
                + "."
            )

        # Check the inputs and Jacobians consistency
        func_names = self.outvars + self.opt_problem.get_constraints_names()
        PostOptimalAnalysis._check_jacobians(functions_jac, func_names, inputs)

        # Compute the Lagrange multipliers
        self._compute_lagrange_multipliers()

        # Compute the Jacobian of the Lagrangian
        jac = self.compute_lagrangian_jac(functions_jac, inputs)

        return jac

    @staticmethod
    def _check_jacobians(functions_jac, func_names, inputs):
        """Checks the consistency of the Jacobians with the required inputs.

        :param functions_jac: Jacobians of the optimization function w.r.t. the
            differentiation inputs
        :type functions_jac: dict(dict(ndarray))
        :param func_names: names list of the functions differentiated
        :type func_names: list(str)
        :param inputs: names list of the inputs w.r.t. which to differentiate
        :type inputs: list(str)
        """
        # Check the consistency of the Jacobians
        for output_name in func_names:
            jac_out = functions_jac.get(output_name)
            if jac_out is None:
                raise ValueError("Jacobian of " + output_name + " is missing.")
            for input_name in inputs:
                jac_block = jac_out.get(input_name)
                if jac_block is None:
                    raise ValueError(
                        "Jacobian of "
                        + output_name
                        + " with respect to "
                        + input_name
                        + " is missing."
                    )
                if not isinstance(jac_block, ndarray):
                    raise ValueError(
                        "Jacobian of "
                        + output_name
                        + " with respect to "
                        + input_name
                        + " must be of type ndarray."
                    )
                if len(jac_block.shape) != 2:
                    raise ValueError(
                        "Jacobian of "
                        + output_name
                        + " with respect to "
                        + input_name
                        + " must be a 2-dimensional ndarray."
                    )

    def _compute_lagrange_multipliers(self):
        """Computes the Lagrange multipliers at the solution."""
        self.lagrange_computer.compute(self.x_opt, self.ineq_tol)

    def compute_lagrangian_jac(self, functions_jac, inputs):
        """Computes the Jacobian of the Lagrangian.

        :param functions_jac: Jacobians of the optimization function w.r.t. the
            differentiation inputs
        :type functions_jac: dict(dict(ndarray))
        :param inputs: names list of the inputs w.r.t. which to differentiate
        :type inputs: list(str)
        """
        # Get the Lagrange multipliers
        multipliers = self.lagrange_computer.lagrange_multipliers
        if multipliers is None:
            self._compute_lagrange_multipliers()
            multipliers = self.lagrange_computer.lagrange_multipliers
        _, mul_ineq = multipliers.get(LagrangeMultipliers.INEQUALITY, ([], []))
        _, mul_eq = multipliers.get(LagrangeMultipliers.EQUALITY, ([], []))

        # Build the Jacobians of the active constraints
        act_ineq_jac = self._get_act_ineq_jac(functions_jac, inputs)
        eq_jac = self._get_eq_jac(functions_jac, inputs)

        jac = {self.outvars[0]: dict(), self.MULT_DOT_CONSTR_JAC: dict()}
        for input_name in inputs:

            # Contribution of the objective
            jac_obj_arr = functions_jac[self.outvars[0]][input_name]
            jac_cstr_arr = zeros_like(jac_obj_arr)

            # Contributions of the inequality constraints
            jac_ineq_arr = act_ineq_jac.get(input_name)
            if jac_ineq_arr is not None:
                jac_cstr_arr += dot(mul_ineq, jac_ineq_arr)

            # Contributions of the equality constraints
            jac_eq_arr = eq_jac.get(input_name)
            if jac_eq_arr is not None:
                jac_cstr_arr += dot(mul_eq, jac_eq_arr)

            # Assemble the Jacobian of the Lagrangian
            if not self.opt_problem.minimize_objective:
                jac_cstr_arr *= -1.0
            jac[self.MULT_DOT_CONSTR_JAC][input_name] = jac_cstr_arr
            jac[self.outvars[0]][input_name] = jac_obj_arr + jac_cstr_arr

        return jac

    def _get_act_ineq_jac(self, jacobians, inputs):
        """Builds the Jacobian of the active inequality constraints.

        :param jacobians: Jacobians of the inequality constraints w.r.t. the
            differentiation inputs
        :type jacobians: dict(dict(ndarray))
        :param inputs: names list of the differentiation inputs
        :type inputs: list(str)
        """
        # Get the active constraints
        ineq_cstr = self.opt_problem.get_active_ineq_constraints(
            self.x_opt, self.ineq_tol
        )

        # Build the Jacobian
        jac_dict = dict()
        for input_name in inputs:
            jac_input_list = []
            for func, act_set in ineq_cstr.items():
                if True in act_set:
                    jac_block = jacobians[func.name][input_name]
                    jac_block = jac_block[atleast_1d(act_set), :]
                    jac_input_list.append(jac_block)
            if jac_input_list:
                jac_input_arr = vstack(jac_input_list)
                jac_dict[input_name] = jac_input_arr

        return jac_dict

    def _get_eq_jac(self, jacobians, inputs):
        """Builds the Jacobian of the equality constraints.

        :param jacobians: Jacobians of the equality constraints w.r.t. the
            differentiation inputs
        :type jacobians: dict(dict(ndarray))
        :param inputs: names list of the differentiation inputs
        :type inputs: list(str)
        """
        eq_cstr = self.opt_problem.get_eq_constraints()
        jac_dict = dict()
        for input_name in inputs:
            jac_input_list = [jacobians[func.name][input_name] for func in eq_cstr]
            if jac_input_list:
                jac_input_arr = vstack(jac_input_list)
                jac_dict[input_name] = jac_input_arr

        return jac_dict



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#    initial documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
The aerostructure MDO problem
*****************************

The **aerostructure** module implements all :class:`.MDODiscipline`
included in the Aerostructure problem:

.. math::

   \text{OVERALL AIRCRAFT DESIGN} = \left\{
   \begin{aligned}
   &\text{minimize }\text{range}(\text{thick\_airfoils},
   \text{thick\_panels}, \text{sweep}) = 8
   \times10^{11}\times\text{lift}\times\text{mass}/\text{drag} \\
   &\text{with respect to }\text{thick\_airfoils},\,\text{thick\_panels},
   \,\text{sweep} \\
   &\text{subject to }\\
   & \text{rf}-0.5 = 0\\
   & \text{lift}-0.5 \leq 0
   \end{aligned}\right.

where

.. math::

       \text{AERODYNAMICS} = \left\{
           \begin{aligned}
        &\text{drag}=0.1\times((\text{sweep}/360)^2 + 200 +
        \text{thick\_airfoils}^2 - \text{thick\_airfoils} -
         4\times\text{displ})\\
        &\text{forces}=10\times\text{sweep} +
        0.2\times\text{thick\_airfoils}-0.2\times\text{displ}\\
        &\text{lift}=(\text{sweep} + 0.2\times\text{thick\_airfoils}-
        2\times\text{displ})/3000
           \end{aligned}
           \right.

and

.. math::

       \text{STRUCTURE} = \left\{
           \begin{aligned}
        &\text{mass}=4000\times(\text{sweep}/360)^3 + 200000 +
        100\times\text{thick\_panels} + 200\times\text{forces}\\
        &\text{rf}=3\times\text{sweep} - 6\times\text{thick\_panels} +
        0.1\times\text{forces} + 55\\
        &\text{displ}=2\times\text{sweep} + 3\times\text{thick\_panels} -
        2\times\text{forces}
           \end{aligned}
           \right.



"""
from __future__ import division, unicode_literals

from numpy import array, atleast_2d, complex128, ones

from gemseo.core.discipline import MDODiscipline


def get_inputs(names=None):
    """Generate initial solution.

    :param names: input names (Default value = None)
    :type names: list(str)
    """
    inputs = {
        "drag": ones(1, dtype=complex128),
        "forces": ones(1, dtype=complex128),
        "lift": ones(1, dtype=complex128),
        "mass": ones(1, dtype=complex128),
        "displ": ones(1, dtype=complex128),
        "sweep": ones(1, dtype=complex128),
        "thick_airfoils": ones(1, dtype=complex128),
        "thick_panels": ones(1, dtype=complex128),
        "reserve_fact": ones(1, dtype=complex128),
    }
    if names is None:
        return inputs
    return {name: inputs.get(name) for name in names}


class Mission(MDODiscipline):
    """**Aerostructure** is the :class:`.MDODiscipline`
    implementing the computation of the Aerostructure's objective
    and constraints discipline."""

    def __init__(self, r_val=0.5, lift_val=0.5):
        """Constructor."""
        super(Mission, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = get_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        self.r_val = r_val
        self.lift_val = lift_val

    def _run(self):
        """Defines the execution of the process, given that data has been checked.

        Compute the outputs (= objective value and constraints at system level) of the
        aerostructure analytical problem.
        """
        lift, mass, drag, reserve_fact = self.get_inputs_by_name(
            ["lift", "mass", "drag", "reserve_fact"]
        )
        obj = array([self.compute_range(lift, mass, drag)], dtype=complex128)
        c_lift = array([self.c_lift(lift, self.lift_val)], dtype=complex128)
        c_rf = array([self.c_rf(reserve_fact)], dtype=complex128)
        self.store_local_data(range=obj, c_lift=c_lift, c_rf=c_rf)

    @staticmethod
    def compute_range(lift, mass, drag):
        """Objective function: :math:`range=8.10^{11}*lift/(mass*drag)`

        :param lift: lift
        :type lift: ndarray
        :param mass: mass
        :type mass: ndarray
        :param drag: drag
        :type drag: ndarray
        :returns: Objective value (range)
        :rtype: float
        """
        return 8e11 * lift[0] / (mass[0] * drag[0])

    @staticmethod
    def c_lift(lift, lift_val=0.5):
        """lift constraint: :math:`lift-0.5`

        :param lift: lift
        :type lift: ndarray
        :returns: Value of the lift constraint
        """
        return lift[0] - lift_val

    @staticmethod
    def c_rf(reserve_fact, rf_val=0.5):
        """reserve_fact constraint: :math:`rf-0.5`

        :param reserve_fact: reserve_fact
        :type reserve_fact: ndarray
        :param rf_val: rf_val
        :type rf_val: ndarray
        :returns: Value of the rf constraint
        """
        return reserve_fact[0] - rf_val

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Computes the jacobian of [range, c_rf, c_lift] w.r.t [lift, mass, drag, rf]

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        # Initialize all matrices to zeros
        self._init_jacobian(inputs, outputs, with_zeros=True)
        drag, lift, mass = self.get_inputs_by_name(["drag", "lift", "mass"])
        self.jac["c_lift"]["lift"] = ones((1, 1))
        self.jac["c_rf"]["reserve_fact"] = ones((1, 1))
        self.jac["range"]["lift"] = atleast_2d(array([8e11 / (mass[0] * drag[0])]))
        self.jac["range"]["mass"] = atleast_2d(
            array([-8e11 * lift[0] / (mass[0] ** 2 * drag[0])])
        )
        self.jac["range"]["drag"] = atleast_2d(
            array([-8e11 * lift[0] / (mass[0] * drag[0] ** 2)])
        )


class Aerodynamics(MDODiscipline):
    """**Aerodynamics** is the :class:`.MDODiscipline`
    implementing the computation of the Aerodynamics' equations:
    [drag, forces, lift] = f(sweep, thick_airfoils, displ)."""

    def __init__(self):
        super(Aerodynamics, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = get_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Defines the execution of the process, given that data has been checked.

        Solve a coupling equation in functional form and compute coupling variable
        """
        sweep, thick_airfoils, displ = self.get_inputs_by_name(
            ["sweep", "thick_airfoils", "displ"]
        )
        drag_out = array([self.compute_drag(sweep, thick_airfoils, displ)])
        lift_out = array([self.compute_lift(sweep, thick_airfoils, displ)])
        forces_out = array([self.compute_forces(sweep, thick_airfoils, displ)])
        self.store_local_data(drag=drag_out, forces=forces_out, lift=lift_out)

    @staticmethod
    def compute_drag(sweep, thick_airfoils, displ):
        """Compute the coupling
        :math:`drag=0.1*((sweep/360)^2 + 200 + thick_airfoils^2
        - thick_airfoils - 4*displ)`

        :param sweep: sweep
        :type sweep: ndarray
        :param thick_airfoils: thick_airfoils
        :type thick_airfoils: ndarray
        :param displ: displ
        :type displ: ndarray
        :returns: drag
        :rtype: float
        """
        return 0.1 * (
            (sweep[0] / 360) ** 2
            + 200
            + thick_airfoils[0] ** 2
            - thick_airfoils[0]
            - 4 * displ[0]
        )

    @staticmethod
    def compute_forces(sweep, thick_airfoils, displ):
        """Compute the coupling
        :math:`forces=10*sweep + 0.2*thick_airfoils-0.2*displ`

        :param sweep: sweep
        :type sweep: ndarray
        :param thick_airfoils: thick_airfoils
        :type thick_airfoils: ndarray
        :param displ: displ
        :type displ: ndarray
        :returns: drag
        :rtype: float
        """
        return 10 * sweep[0] + 0.2 * thick_airfoils[0] - 0.2 * displ[0]

    @staticmethod
    def compute_lift(sweep, thick_airfoils, displ):
        """Compute the coupling
        :math:`lift=(sweep + 0.2*thick_airfoils-2.*displ)/3000.`

        :param sweep: sweep
        :type sweep: ndarray
        :param thick_airfoils: thick_airfoils
        :type thick_airfoils: ndarray
        :param displ: displ
        :type displ: ndarray
        :returns: drag
        :rtype: float
        """
        return (sweep[0] + 0.2 * thick_airfoils[0] - 2.0 * displ[0]) / 3000.0

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the Jacobian of [drag, forces, lift]

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        # Initialize all matrices to zeros
        self._init_jacobian(inputs, outputs, with_zeros=True)
        sweep, thick_airfoils = self.get_inputs_by_name(["sweep", "thick_airfoils"])
        self.jac["drag"]["sweep"] = atleast_2d(
            array([0.1 * 2.0 * sweep[0] / 360.0 ** 2.0])
        )
        self.jac["drag"]["thick_airfoils"] = atleast_2d(
            array([0.1 * (2.0 * thick_airfoils[0] - 1.0)])
        )
        self.jac["drag"]["displ"] = atleast_2d(0.1 * array([-4.0]))
        self.jac["forces"]["sweep"] = atleast_2d(array([10.0]))
        self.jac["forces"]["thick_airfoils"] = atleast_2d(array([0.2]))
        self.jac["forces"]["displ"] = atleast_2d(array([-0.2]))
        self.jac["lift"]["sweep"] = atleast_2d(array([1.0 / 3000.0]))
        self.jac["lift"]["thick_airfoils"] = atleast_2d(array([0.2 / 3000.0]))
        self.jac["lift"]["displ"] = atleast_2d(array([-2.0 / 3000.0]))


class Structure(MDODiscipline):
    """**Structure** is the :class:`.MDODiscipline`
    implementing the computation of the Structure' equations:
    [mass, rf, displ] = f(sweep, thick_panels, forces)"""

    def __init__(self):
        super(Structure, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = get_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Defines the execution of the process, given that data has been checked.

        Solve a coupling equation in functional form and compute coupling variables.
        """
        sweep, thick_panels, forces = self.get_inputs_by_name(
            ["sweep", "thick_panels", "forces"]
        )
        mass_out = array([self.compute_mass(sweep, thick_panels, forces)])
        rf_out = array([self.compute_rf(sweep, thick_panels, forces)])
        displ_out = array([self.compute_displ(sweep, thick_panels, forces)])
        self.store_local_data(mass=mass_out, reserve_fact=rf_out, displ=displ_out)

    @staticmethod
    def compute_mass(sweep, thick_panels, forces):
        """Compute the coupling
        :math:`mass=4000*(sweep/360)^3 + 200000 + 100*thick_panels
        + 200.0*forces`

        :param sweep: sweep
        :type sweep: ndarray
        :param thick_panels: thick_panels
        :type thick_panels: ndarray
        :param forces: forces
        :type forces: ndarray
        :returns: mass
        :rtype: float
        """
        return (
            4000 * (sweep[0] / 360) ** 3
            + 200000
            + 100 * thick_panels[0]
            + 200.0 * forces[0]
        )

    @staticmethod
    def compute_rf(sweep, thick_panels, forces):
        """Compute the coupling.

        :math:`rf=-3*sweep - 6*thick_panels + 0.1*forces + 55`

        :param sweep: sweep
        :type sweep: ndarray
        :param thick_panels: thick_panels
        :type thick_panels: ndarray
        :param forces: forces
        :type forces: ndarray
        :returns: RF
        :rtype: float
        """
        return -3 * sweep[0] - 6 * thick_panels[0] + 0.1 * forces[0] + 55

    @staticmethod
    def compute_displ(sweep, thick_panels, forces):
        """Compute the coupling.

        :math:`displ=2*sweep + 3*thick_panels - 2.*forces`

        :param sweep: sweep
        :type sweep: ndarray
        :param thick_panels: thick_panels
        :type thick_panels: ndarray
        :param forces: forces
        :type forces: ndarray
        :returns: displ
        :rtype: float
        """
        return 2 * sweep[0] + 3 * thick_panels[0] - 2.0 * forces[0]

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the Jacobian of [mass, rf, displ]

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        # Initialize all matrices to zeros
        self._init_jacobian(inputs, outputs, with_zeros=True)
        sweep = self.get_inputs_by_name("sweep")
        self.jac["mass"]["sweep"] = atleast_2d(
            array([4000.0 * 3.0 * sweep[0] ** 2 / 360.0 ** 3])
        )
        self.jac["mass"]["thick_panels"] = atleast_2d(array([100.0]))
        self.jac["mass"]["forces"] = atleast_2d(array([200.0]))
        self.jac["reserve_fact"]["sweep"] = atleast_2d(array([-3.0]))
        self.jac["reserve_fact"]["thick_panels"] = atleast_2d(array([-6.0]))
        self.jac["reserve_fact"]["forces"] = atleast_2d(array([0.1]))
        self.jac["displ"]["sweep"] = atleast_2d(array([2.0]))
        self.jac["displ"]["thick_panels"] = atleast_2d([3.0])
        self.jac["displ"]["forces"] = atleast_2d(array([-2.0]))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable problem
================
"""
from __future__ import division, unicode_literals

import logging

from gemseo.algos.parameter_space import ParameterSpace
from gemseo.problems.scalable.parametric.core.problem import TMProblem
from gemseo.problems.scalable.parametric.core.variables import get_u_local_name
from gemseo.problems.scalable.parametric.disciplines import (
    TMMainDiscipline,
    TMSubDiscipline,
)

LOGGER = logging.getLogger(__name__)

MDA_TOLERANCE = {"tolerance": 1e-14, "linear_solver_tolerance": 1e-14}
ALGO_OPTIONS = {
    "xtol_rel": 1e-4,
    "ftol_rel": 1e-4,
    "xtol_abs": 1e-4,
    "ftol_abs": 1e-4,
    "ineq_tolerance": 1e-3,
    "eq_tolerance": 1e-3,
}


OBJECTIVE_NAME = "obj"

COUPLING_DIR = "coupling"
COEFF_DIR = "coefficients"
OPTIM_DIR = "opthistoryview"


class TMScalableProblem(TMProblem):

    """The scalable problem from Tedford and Martins, 2010, builds a list of strongly
    coupled scalable disciplines completed by a system discipline computing the
    objective function and the constraints.

    These disciplines are defined on an unit design space (parameters comprised in [0,
    1]).
    """

    @classmethod
    def _create_main_model(cls, c_constraint, default_inputs):
        """Create main model.

        :param ndarray c_constraint: coefficients for constraint.
        :param dict(ndarray) default_inputs: default inputs.
        """
        return TMMainDiscipline(c_constraint, default_inputs)

    @classmethod
    def _create_sub_model(cls, index, c_shared, c_local, c_cpl, default_inputs):
        """Create sub model.

        :param int index: model index.
        :param ndarray c_shared: coefficients for shared design parameters.
        :param ndarray c_local: coefficients for local design parameters.
        :param ndarray c_cpl: coefficients for coupling variables.
        :param dict(ndarray) default_inputs: default inputs.
        """
        return TMSubDiscipline(index, c_shared, c_local, c_cpl, default_inputs)

    @property
    def disciplines(self):
        """Alias for self.models."""
        return self.models

    @property
    def n_disciplines(self):
        """Alias for self.n_submodels."""
        return self.n_submodels

    @property
    def main_discipline(self):
        """Main disciplines.

        :return: main discipline.
        :rtype: TMDiscipline
        """
        return self.models[0]

    @property
    def sub_disciplines(self):
        """Sub-disciplines.

        :return: list of disciplines.
        :rtype: list(TMDiscipline)
        """
        return self.models[1:]

    def get_design_space(self):
        """Get the TM design space.

        :return: instance of the design space
        :rtype: DesignSpace
        """
        d_s = super(TMScalableProblem, self).get_design_space()
        design_space = ParameterSpace()
        for name in d_s.names:
            size = d_s.sizes[name]
            l_b = d_s.lower_bounds[name]
            u_b = d_s.upper_bounds[name]
            value = d_s.default_values[name]
            design_space.add_variable(name, size, "float", l_b, u_b, value)
        if self._noised_coupling:
            for index in range(self.n_submodels):
                name = get_u_local_name(index)
                size = self.n_coupling
                distribution = "OTNormalDistribution"
                design_space.add_random_variable(name, distribution, size[index])
        return design_space

    def reset_disciplines(self):
        """Reset the disciplines, setting n_calls=0, n_calls_linearize=0, exec_time=0
        and local_data={}."""
        for discipline in self.disciplines:
            discipline.n_calls = 0
            discipline.exec_time = 0.0
            discipline.n_calls_linearize = 0
            discipline.local_data = {}



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Draw parallel coordinates from a :class:`.Dataset`.

The :class:`.ParallelCoordinates` class implements the parallel coordinates
plot, a.k.a. cowebplot, which is a way to visualize :math:`n` samples of a
high-dimensional vector

.. math::

   x=(x_1,x_2,\ldots,x_d)\in\mathbb{R}^d

in a 2D referential by representing each sample

.. math::

   x^{(i)}=(x_1^{(i)},x_2^{(i)},\ldots,x_d^{(i)})

as a piece-wise line where the x-values of the nodes from left to right
are the values of :math:`x_1`, :math:`x_2`, ... and :math:`x_d^{(i)}`.

A variable name is required by the :meth:`.DatasetPlot.execute` method
by means of the :code:`classifier` keyword in order to color the curves
according to the value of the variable name. This is useful when the data is
labeled or when we are looking for the samples for which the classifier value
is comprised in some interval specified by the :code:`lower` and :code:`upper`
arguments
(default values are set to :code:`-inf` and :code:`inf` respectively).
In the latter case, the color scale is composed of only two values: one for
the samples positively classified and one for the others.
"""
from __future__ import division, unicode_literals

from typing import List, Mapping

from matplotlib.figure import Figure
from numpy import inf
from pandas.plotting import parallel_coordinates

from gemseo.post.dataset.dataset_plot import DatasetPlot


class ParallelCoordinates(DatasetPlot):
    """Parallel coordinates."""

    def _plot(
        self,
        properties,  # type: Mapping
        classifier,  # type: str
        lower=-inf,  # type: float
        upper=inf,  # type: float
        **kwargs
    ):  # type: (...) -> List[Figure]
        """
        Args:
            classifier: The name of the variable to group the data.
            lower: The lower bound of the cluster.
            upper: The upper bound of the cluster.
        """
        if classifier not in self.dataset.variables:
            raise ValueError(
                "Classifier must be one of these names: "
                + ", ".join(self.dataset.variables)
            )
        label, varname = self._get_label(classifier)
        dataframe = self.dataset.export_to_dataframe()
        cluster = varname
        columns = list(dataframe.columns)

        def is_btw(row):
            return lower < row[varname] < upper

        if lower != -inf or upper != inf:
            cluster = "{} < {} < {}".format(lower, label, upper)
            cluster = ("classifiers", cluster, "0")
            dataframe[cluster] = dataframe.apply(is_btw, axis=1)
        axes = parallel_coordinates(dataframe, cluster, cols=columns, **kwargs)
        axes.set_xticklabels(self._get_variables_names(columns))
        if lower != -inf or upper != inf:
            default_title = "Cobweb plot based on the classifier: {}".format(cluster[1])
        else:
            default_title = None
            axes.get_legend().remove()
        axes.set_title(self.title or default_title)
        return [axes.figure]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Benoit Pauwels
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""An adapter from `.OptimizationProblem` to a pSeven ProblemGeneric."""

from enum import Enum
from typing import Dict, List, Mapping, Optional, Tuple, Union

from da import p7core
from numpy import array, atleast_1d, concatenate, full, full_like, ndarray

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.algos.stop_criteria import TerminationCriterion
from gemseo.core.function import MDOFunction, MDOLinearFunction, MDOQuadraticFunction


class CostType(Enum):
    """The evaluation cost type of pSeven functions."""

    CHEAP = "Cheap"
    EXPENSIVE = "Expensive"


class PSevenProblem(p7core.gtopt.ProblemGeneric):
    """Adapter of OptimizationProblem to da.p7core.gtopt.ProblemGeneric.

    The methods prepare_problem() and evaluate() are defined according to pSeven's
    requirements. Refer to the API documentation of pSeven Core for more information.
    """

    def __init__(
        self,
        problem,  # type: OptimizationProblem
        evaluation_cost_type=None,  # type: Optional[Mapping[str, CostType]]
        expensive_evaluations=None,  # type: Optional[Mapping[str, int]]
        lower_bounds=None,  # type: Optional[ndarray]
        upper_bounds=None,  # type: Optional[ndarray]
        initial_point=None,  # type: Optional[ndarray]
    ):  # type: (...) -> None
        """# noqa:D205,D212,D415
        Args:
            problem: The optimization problem to be adapted to pSeven.
            evaluation_cost_type: The evaluation cost type of each function of the
                problem.
                If None, the evaluation cost types default to "Cheap".
            expensive_evaluations: The maximal number of expensive evaluations for
                each function of the problem.
                If None, this number is set automatically by pSeven.
            lower_bounds: The lower bounds on the design variables.
                If None, the lower bounds are read from the design space.
            upper_bounds: The upper bounds on the design variables.
                If None, the upper bounds are read from the design space.
            initial_point: The initial values of the design variables.
                If None, the initial values are read from the design space.
        """
        self.__problem = problem

        if evaluation_cost_type is None:
            self.__evaluation_cost_type = dict()
        else:
            self.__evaluation_cost_type = evaluation_cost_type

        if expensive_evaluations is None:
            self.__expensive_evaluations = dict()
        else:
            self.__expensive_evaluations = expensive_evaluations

        # Set the design variables bounds and initial values
        design_space = problem.design_space

        if lower_bounds is None:
            self.__lower_bounds = design_space.get_lower_bounds()
        else:
            self.__lower_bounds = lower_bounds

        if upper_bounds is None:
            self.__upper_bounds = design_space.get_upper_bounds()
        else:
            self.__upper_bounds = upper_bounds

        if initial_point is not None:
            self.__initial_point = initial_point
        elif design_space.has_current_x():
            self.__initial_point = design_space.get_current_x()
        else:
            self.__initial_point = full(design_space.dimension, None)

    def prepare_problem(self):  # type: (...) -> None
        """Initialize the problem for pSeven."""
        self.__add_variables()
        self.__add_objectives()
        self.__add_constraints()

    def __add_variables(self):  # type: (...) -> None
        """Add the design variables to the pSeven problem."""
        design_space = self.__problem.design_space
        for var_name in design_space.variables_names:
            var_indexes = design_space.get_variables_indexes([var_name])
            lower_bound = self.__lower_bounds[var_indexes]
            upper_bound = self.__upper_bounds[var_indexes]
            current_x = self.__initial_point[var_indexes]
            indexed_names = design_space.get_indexed_var_name(var_name)
            for index in range(len(indexed_names)):
                bounds = (lower_bound[index], upper_bound[index])
                initial_guess = current_x[index]
                pseven_type = self.__get_p7_variable_type(var_name, index)
                hints = {"@GT/VariableType": pseven_type}
                self.add_variable(bounds, initial_guess, indexed_names[index], hints)

    def __add_objectives(self):  # type: (...) -> None
        """Add the objectives to the pSeven problem."""
        objective = self.__problem.objective
        hints = self.__get_p7_function_hints(objective)
        dimension = self.__get_function_dimension(objective)

        if dimension > 1:
            for index in range(dimension):
                name = self.__get_component_name(objective, index)
                self.add_objective(name, hints)
        else:
            self.add_objective(objective.name, hints)

        # Add the objectives gradients
        if objective.has_jac():
            self.enable_objectives_gradient()

    def __add_constraints(self):  # type: (...) -> None
        """Add the constraints to the pSeven problem."""
        problem = self.__problem

        for constraint in problem.constraints:
            bounds = self.__get_p7_constraint_bounds(constraint)
            hints = self.__get_p7_function_hints(constraint)
            dimension = self.__get_function_dimension(constraint)
            if dimension > 1:
                for index in range(dimension):
                    name = self.__get_component_name(constraint, index)
                    self.add_constraint(bounds, name, hints)
            else:
                self.add_constraint(bounds, constraint.name, hints)

        # Add the constraints gradients
        differentiable = all(constraint.has_jac() for constraint in problem.constraints)
        if problem.has_constraints() and differentiable:
            self.enable_constraints_gradient()

    def __get_p7_variable_type(
        self,
        variable_name,  # type: str
        index,  # type: int
    ):  # type: (...) -> str
        """Return the pSeven variable type associated with a design variable component.

        Args:
            variable_name: The name of the design variable.
            index: The index of the variable component.

        Returns:
            The pSeven variable type.

        Raises:
            TypeError: If the type of the design variable is not supported by pSeven.
        """
        var_type = self.__problem.design_space.get_type(variable_name)[index]
        if var_type == DesignSpace.FLOAT:
            return "Continuous"
        if var_type == DesignSpace.INTEGER:
            return "Integer"
        raise TypeError("Unsupported design variable type: {}".format(var_type))
        # TODO: For future reference, pSeven also supports discrete and categorical
        #  variables.

    def __get_p7_function_hints(
        self,
        function,  # type: MDOFunction
    ):  # type: (...) -> Dict[str, Union[str, int]]
        """Return the pSeven hints associated with a function.

        Args:
            function: The function.

        Returns:
            The pSeven hints.
        """
        linearity_type = PSevenProblem.__get_p7_linearity_type(function)
        hints = {"@GTOpt/LinearityType": linearity_type}
        name = function.name
        if name in self.__evaluation_cost_type:
            hints["@GTOpt/EvaluationCostType"] = self.__evaluation_cost_type[name].value
        if name in self.__expensive_evaluations:
            hints["@GTOpt/ExpensiveEvaluations"] = self.__expensive_evaluations[name]
        return hints

    @staticmethod
    def __get_p7_linearity_type(
        function,  # type: MDOFunction
    ):  # type: (...) -> str
        """Return the pSeven linearity type of a function.

        Args:
            function: The function.

        Returns:
            The pSeven linearity type.
        """
        if isinstance(function, MDOLinearFunction):
            return "Linear"
        if isinstance(function, MDOQuadraticFunction):
            return "Quadratic"
        return "Generic"

    def __get_function_dimension(
        self,
        function,  # type: MDOFunction
    ):  # type: (...) -> int
        """Return the dimension of a function.

        Args:
            function: The function.

        Returns:
            The dimension of the function.

        Raises:
            RuntimeError: If the function dimension is unavailable.
        """
        design_space = self.__problem.design_space
        if function.has_dim():
            return function.dim
        elif design_space.has_current_x():
            current_x = design_space.get_current_x()
            return atleast_1d(function(current_x)).size
        else:
            raise RuntimeError("The function dimension is not available.")

    @staticmethod
    def __get_component_name(
        function,  # type: MDOFunction
        index,  # type: int
    ):  # type: (...) -> str
        """Return the name of function component.

        Args:
            function: The function.
            index: The index of the function component.

        Returns:
            The name of the function component.
        """
        return "{}{}{}".format(function.name, DesignSpace.SEP, index)

    @staticmethod
    def __get_p7_constraint_bounds(
        constraint,  # type: MDOFunction
    ):  # type: (...) -> Tuple[Optional[float], float]
        """Return the pSeven bounds associated with a constraint.

        Args:
            constraint: The constraint.

        Returns:
            The lower bound, the upper bound.

        Raises:
            ValueError: If the constraint type is invalid.
        """
        if constraint.f_type == MDOFunction.TYPE_EQ:
            return 0.0, 0.0
        if constraint.f_type == MDOFunction.TYPE_INEQ:
            return None, 0.0
        raise ValueError("Invalid constraint type.")
        # TODO: For future reference, pSeven support constraints bounded from both
        #  sides.

    def evaluate(
        self,
        queryx,  # type: ndarray
        querymask,  # type: ndarray
    ):  # type: (...) -> Tuple[List[List[float]], List[ndarray]]
        """Compute the values of the objectives and the constraints for pSeven.

        Args:
            queryx: The points to evaluate. (2D-array where each row is a point.)
            querymask: The evaluation request mask.

        Returns:
            The evaluation result. (2D-array-like with one row per point.),
            The evaluation masks. (Idem.)

        Raises:
            p7core.UserTerminated: If a termination criterion is reached.
        """
        functions_batch = list()
        output_masks_batch = list()
        for x, mask in zip(queryx, querymask):
            # Evaluate the functions, unless a stopping criterion is satisfied
            functions = list()
            try:
                # Compute the objectives values
                objectives, obj_mask = self.__compute_objectives(x, mask)
                functions.extend(objectives)
                # Compute the constraints values
                constraints, constr_mask = self.__compute_constraints(
                    x, mask[len(functions) :]
                )
                functions.extend(constraints)
                # Compute the objectives gradients
                obj_grads, obj_grads_mask = self.__compute_objectives_gradients(
                    x, mask[len(functions) :]
                )
                functions.extend(obj_grads)
                # Compute the constraints gradients
                constr_grads, constr_grads_mask = self.__compute_constraints_gradients(
                    x, mask[len(functions) :]
                )
            except TerminationCriterion:
                # Interrupt pSeven
                raise p7core.UserTerminated("Gemseo stopping criterion satisfied")
            else:
                functions.extend(constr_grads)
                functions_batch.append(functions)
                output_mask = concatenate(
                    [obj_mask, constr_mask, obj_grads_mask, constr_grads_mask]
                )
                output_masks_batch.append(output_mask)

        return functions_batch, output_masks_batch

    def __compute_objectives(
        self,
        x_vec,  # type: ndarray
        mask,  # type: ndarray
    ):  # type: (...) -> Tuple[List[float], ndarray]
        obj_dim = self.__get_function_dimension(self.__problem.objective)

        if True in mask[:obj_dim]:
            objectives = atleast_1d(self.__problem.objective(x_vec)).tolist()
            output_mask = full_like(mask[:obj_dim], True)
        else:
            objectives = [None] * obj_dim
            output_mask = full_like(mask[:obj_dim], False)

        return objectives, output_mask

    def __compute_constraints(
        self,
        x_vec,  # type: ndarray
        mask,  # type: ndarray
    ):  # type: (...) -> Tuple[List[float], ndarray]
        constraints = list()
        output_mask = list()
        n_inds = 0

        for constraint in self.__problem.constraints:
            constr_dim = self.__get_function_dimension(constraint)
            if True in mask[n_inds : n_inds + constr_dim]:
                constraints.extend(atleast_1d(constraint(x_vec)).tolist())
                output_mask.extend([True] * constr_dim)
            else:
                constraints.extend([None] * constr_dim)
                output_mask.extend([False] * constr_dim)
            n_inds += constr_dim

        return constraints, array(output_mask)

    def __compute_objectives_gradients(
        self,
        x_vec,  # type: ndarray
        mask,  # type: ndarray
    ):  # type: (...) -> Tuple[List[float], ndarray]
        pb_dim = self.__problem.dimension
        obj_dim = self.__get_function_dimension(self.__problem.objective)
        n_values = obj_dim * pb_dim

        if self.__problem.objective.has_jac() and True in mask[:n_values]:
            obj_grads = self.__problem.objective.jac(x_vec).flatten().tolist()
            output_mask = full_like(mask[:n_values], True)
        elif self.__problem.objective.has_jac():
            obj_grads = [None] * n_values
            output_mask = full_like(mask[:n_values], False)
        else:
            obj_grads = []
            output_mask = array([])

        return obj_grads, output_mask

    def __compute_constraints_gradients(
        self,
        x_vec,  # type: ndarray
        mask,  # type: ndarray
    ):  # type: (...) -> Tuple[List[float], ndarray]
        constr_grads = list()
        output_mask = list()
        pb_dim = self.__problem.dimension
        n_inds = 0

        if all(constraint.has_jac() for constraint in self.__problem.constraints):
            for constraint in self.__problem.constraints:
                constr_dim = self.__get_function_dimension(constraint)
                n_values = constr_dim * pb_dim
                if True in mask[n_inds : n_inds + n_values]:
                    constr_grads.extend(constraint.jac(x_vec).flatten().tolist())
                    output_mask.extend([True] * n_values)
                else:
                    constr_grads.extend([None] * n_values)
                    output_mask.extend([False] * n_values)
                n_inds += n_values

        return constr_grads, array(output_mask)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Decorators to lock and protect critical code sections
*****************************************************
"""
from __future__ import division, unicode_literals

import functools


def synchronized(wrapped):
    """A synchronization decorator to avoid concurrent access of critical sections.

    The wrapped function must be a method of an object
    with a self.lock attribute

    :param wrapped: function to be protected
    """

    @functools.wraps(wrapped)
    def _wrapper(*args, **kwargs):
        """Definition of the synchronization decorator."""
        with args[0].lock:
            return wrapped(*args, **kwargs)

    return _wrapper


def synchronized_hashes(wrapped):
    """A synchronization decorator to avoid concurrent access of critical sections.

    The wrapped function must be a method of an object
    with a self.lock_hashes attribute

    :param wrapped: function to be protected
    """

    @functools.wraps(wrapped)
    def _wrapper(*args, **kwargs):
        """Definition of the synchronization decorator."""
        with args[0].lock_hashes:
            return wrapped(*args, **kwargs)

    return _wrapper
